from thingsvision.custom_models.resnet50_ecoset import Resnet50_ecoset
from thingsvision.custom_models.vgg16bn_ecoset import VGG16bn_ecoset
from thingsvision.custom_models.alexnet_ecoset import Alexnet_ecoset